/*
 * This file is part of FGGL.
 *
 * FGGL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * FGGL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with FGGL.
 * If not, see <https://www.gnu.org/licenses/>.
 */

#ifndef FGGL_GFX_OGL_SHADER_HPP
#define FGGL_GFX_OGL_SHADER_HPP

#include <cstdio>
#include <fggl/gfx/ogl/backend.hpp>
#include <fggl/data/storage.hpp>

#include <filesystem>
#include <stdexcept>
#include <unordered_map>

namespace fggl::gfx {

	struct ShaderConfig {
		// required
		std::string name;
		std::string vertex;
		std::string fragment;

		// optional parts
		std::string geometry;
		bool hasGeom = false;
	};

	inline ShaderConfig ShaderFromName(const std::string &name) {
		return {
			name,
			name + "_vert.glsl",
			name + "_frag.glsl"
		};
	}

	struct BinaryCache {
		void *data = nullptr;
		GLsizei size = 0;
		GLenum format = GL_INVALID_ENUM;
	};

	class ShaderCache {
		public:
			ShaderCache(std::shared_ptr<fggl::data::Storage> storage);
			~ShaderCache() = default;

			GLuint load(const ShaderConfig &config);
			GLuint getOrLoad(const ShaderConfig &config);

			GLuint get(const std::string &name);


		private:
			std::shared_ptr<fggl::data::Storage> m_storage;
			std::unordered_map<std::string, GLuint> m_shaders;

			// extensions
			void setupIncludes();

			// opengl operations
			bool compileShader(const std::string &, GLuint);

			// file io operations
			bool loadFromDisk(GLuint pid, const ShaderConfig &config);
			void saveToDisk(GLuint pid, const ShaderConfig &config);

			bool m_binary;
			void cacheSave(GLuint pid, BinaryCache *cache);
			bool cacheLoad(GLuint pid, const BinaryCache *cache);
	};

}

#endif
